(* intermediate SML/NG *)
signature INTER_NG =  sig

  datatype inter_ng =
    Seq of inter_ng * inter_ng
  | Txt of string
  | Ws  of string
  | Nd  of (Context.ctxt * inter_ng)

  val go : SMLNG.doc -> SMLNG.doc 
  val fromDoc : SMLNG.doc -> inter_ng
  val toDoc : inter_ng -> SMLNG.doc 

  val sharing_optimization : inter_ng -> inter_ng
  val out  : inter_ng -> TextIO.outstream -> unit

end

structure InterNG :> INTER_NG =  struct
  datatype inter_ng =
    Seq of inter_ng * inter_ng
  | Txt of string
  | Ws  of string
  | Nd  of (Context.ctxt * inter_ng)

  (* context intersection operation *)
  fun ctxt_and(c1:Context.ctxt,c2:Context.ctxt) =
     let
	fun dup Tag 0 = []
	  | dup Tag x = Tag :: dup Tag (x-1)

        val cf =
	(if Context.b(c1) andalso Context.b(c2) then
	                 [SMLNG.B] else []) @
	(if Context.em(c1) andalso Context.em(c2) then
	    [SMLNG.EM] else []) @
	(if Context.i(c1) andalso Context.i(c2) then
	    [SMLNG.I] else []) @
	(if Context.s(c1) andalso Context.s(c2) then
	    [SMLNG.S] else []) @
	(if Context.tt(c1) andalso Context.tt(c2) then
	    [SMLNG.TT] else []) @
	(if Context.u(c1) = Context.u(c2) then
	    dup SMLNG.U (Context.u(c1)) else []) @
	(case (Context.size(c1), Context.size(c2)) of
              (SOME x, SOME y) =>  if x=y then [SMLNG.Size x] else []
          	|   _           => []) @
	(case (Context.color(c1), Context.color(c2)) of
              (SOME x, SOME y) =>  if x=y then [SMLNG.Color x] else []
		|  _           => [])
     in foldl (fn (t,c) => Context.upd(t,c)) Context.init cf
     end

  fun ctxt_sub(c1:Context.ctxt,c2:Context.ctxt)=
     let
	fun dup Tag 0 = []
	  | dup Tag x = Tag :: dup Tag (x-1)

        val cf =
	(if Context.b(c1) andalso not(Context.b(c2)) then
	                 [SMLNG.B] else []) @
	(if Context.em(c1) andalso not(Context.em(c2)) then
	    [SMLNG.EM] else []) @
	(if Context.i(c1) andalso not(Context.i(c2)) then
	    [SMLNG.I] else []) @
	(if Context.s(c1) andalso not(Context.s(c2)) then
	    [SMLNG.S] else []) @
	(if Context.tt(c1) andalso not(Context.tt(c2)) then
	    [SMLNG.TT] else []) @
	(dup SMLNG.U (Context.u(c1)-Context.u(c2))) @
	(case (Context.size(c1), Context.size(c2)) of
              (SOME x, SOME y) =>  if x=y then [] else [SMLNG.Size x]
          	|   _           => []) @
	(case (Context.color(c1), Context.color(c2)) of
              (SOME x, SOME y) =>  if x=y then [] else [SMLNG.Color x]
		|  _           => [])
     in
        foldl (fn (t,c) => Context.upd(t,c)) Context.init cf
     end

  fun gain_by_sharing(c) =
     let
	fun conv b = if b then 1 else 0
	val vsum = (conv (Context.b c)) * 7 +
	           (conv (Context.em c)) * 9 +
		   (conv (Context.i c)) * 7 +
		   (conv (Context.s c)) * 7 +
		   (conv (Context.tt c)) * 9+
		   Context.u c * 7 +
		   (case (Context.size c) of SOME _ => 7 | NONE => 0) +
		   (case (Context.color c) of SOME _ => 7 | NONE => 0)
     in vsum end

  fun sharing_optimization(ng:inter_ng):inter_ng = let
    fun get_weight_list (Seq(Nd(c1,_), tail as (Seq(Nd(c2,_) ,l)))) =
          gain_by_sharing(ctxt_and(c1,c2)) :: get_weight_list tail
    |   get_weight_list (Seq(Nd(c1,_),Nd(c2,_))) = [gain_by_sharing(ctxt_and(c1,c2))]
    |   get_weight_list (Nd(cl,_)) = [0]
    |   get_weight_list _ = raise (Fail "erroneous tree\n")
    val wl=get_weight_list ng

    fun getMax l =
       let
	  fun max [x] n m = (x, n)
	    | max (x::y::xs) n m = if (x > y) then max (x::xs) n (m+1)
				 else max (y::xs) m (m+1)
	    | max [] _ _ = (0,0)
       in max l 1 2
       end

   val (maxv,ps) = getMax wl

   fun transform (Seq (Nd(c1, t1), Seq (Nd(c2,t2), xs)), 1) =
      let
	 val c1_and_c2 = ctxt_and (c1, c2)
	 val c1_sub_c1c2 = ctxt_sub (c1, c1_and_c2)
	 val c2_sub_c1c2 = ctxt_sub (c2, c1_and_c2)
	 val newNode = Nd (c1_and_c2, Seq(Nd(c1_sub_c1c2, t1),
					  Nd(c2_sub_c1c2, t2)))
      in Seq (newNode, xs)
      end
   | transform (Seq (Nd(c1, t1), (Nd(c2,t2))), 1) =
      let
	 val c1_and_c2 = ctxt_and (c1, c2)
	 val c1_sub_c1c2 = ctxt_sub (c1, c1_and_c2)
	 val c2_sub_c1c2 = ctxt_sub (c2, c1_and_c2)
	 val newNode = Nd (c1_and_c2, Seq(Nd(c1_sub_c1c2, t1),
					  Nd(c2_sub_c1c2, t2)))
      in newNode
      end
    | transform (Seq (x,xs), n) = Seq (x, transform (xs, n-1))
    | transform _ = raise (Fail "erroneous tree\n")


  in
    if maxv > 0 then sharing_optimization(transform(ng, ps))
       else ng
  end
  fun dmsg s = TextIO.output(TextIO.stdErr,s)
  (* pretty printer *)
  fun out ng out = let
    fun pr s = TextIO.output(out,s)
    fun f (Seq(d1,d2)) = (f d1;f d2)
    | f (Txt s) = pr s
    | f (Ws s) = pr s
    | f (Nd (c,d)) =
       let
	 val tags = Context.toTags c
       in
	  pr (foldl (fn (x,y) => "<"^(SMLNG.tag2str x) ^ ">" ^ y) "" tags);
          f d;
 	  pr (foldl (fn (x,y) => "</"^(SMLNG.tag2str x) ^ ">" ^ y) "" (List.rev tags))
       end
    in f ng;
       pr "\n"
    end

  fun fromDoc doc = let
    val jlist =
      Context.fold_with_ctxt
      {kSeq=JoinList.append,
       kTxt=(fn (c,s) => JoinList.one (Nd(c,(Txt s)))),
       kWs=(fn (c,s) => JoinList.one (Nd(c,(Ws s)))),
       kNd=(fn (_,t,a) => a)} Context.init doc
    fun loop (JoinList.Nil) = raise (Fail "impossible")
      | loop (JoinList.Cons(x,xs)) =
      case JoinList.pat xs of
	JoinList.Nil => x
      | p => Seq(x,loop p)
  in loop (JoinList.pat jlist)
  end

  structure T = SMLNG
  fun toDoc (Seq(d1,d2)) = T.Seq(toDoc d1,toDoc d2)
    | toDoc (Txt s) = T.Txt s
    | toDoc (Ws s) = T.Ws s
    | toDoc (Nd(ctxt,doc)) = let
	val tags = Context.toTags ctxt
	fun mkNode [] = (toDoc doc)
	  | mkNode (x::xs) = T.Nd(x,mkNode xs)
      in mkNode tags
      end
  fun go d = let
    val d = fromDoc d
(*    val _ = dmsg "\n# cannon\n"
      val _ = out d TextIO.stdErr
*)
    val d = sharing_optimization d
(*    val _ = dmsg "\n#opt\n"
    val _ = out d TextIO.stdErr
*)
    val d = toDoc d
  in d
  end
end
